﻿--
-- Bunker Silo Worker
--
-- @author  LordBanana
-- @date    2023-07-28
--

AIDriveStrategyBunkerSilo = {};
local AIDriveStrategyBunkerSilo_mt = Class(AIDriveStrategyBunkerSilo, AIDriveStrategy);

function AIDriveStrategyBunkerSilo:new(customMt)
    if customMt == nil then
        customMt = AIDriveStrategyBunkerSilo_mt
    end

    local self = AIDriveStrategy:new(customMt)
	setmetatable(self, customMt)

    return self
end

function AIDriveStrategyBunkerSilo:delete()
    AIDriveStrategyBunkerSilo:superClass().delete(self);
end

function AIDriveStrategyBunkerSilo:setAIVehicle(vehicle)

    AIDriveStrategyBunkerSilo:superClass().setAIVehicle(self, vehicle);

    local vehicleSpec = BunkerSiloWorker.getSpec(vehicle)
    vehicle.bunkerSiloWorkerIsStarted = true

	-- First check that we actually wanted to start
	if vehicleSpec.runState == BunkerSiloWorker.STATE_NONE then
		return
	end

    --## set AI direction acccording to current orientation ##--
	
	-- Get directions for vehicle and bunkerSilo
    local dx,_,dz = localDirectionToWorld(vehicle:getAIDirectionNode(), 0, 0, 1)
	local bx,_,bz = localDirectionToWorld(vehicleSpec.bunkerArea.startNode, 0, 0, 1)
	
	-- Get the direction angle in radians for the vehicle
	local vehicleRad = MathUtil.getYRotationFromDirection(dx, dz);
	
	-- Get the direction angle in radians for the bunkerSilo
	local bunkerRad = MathUtil.getYRotationFromDirection(bx, bz);
	
	Debug.log("VehicleRad: " .. vehicleRad .. " bunkerRad: " .. bunkerRad);
	
	-- Check which direction we are closest to. We want two drive in maximum 4 different angles on the
	-- bunker silo.
	-- By subtracting the bunkerRad angle from vehicleRad, we can then use the default snapAIDirection
	-- calculation from vanilla game to decide which angle is closest. Then we just add the bunkerRad
	-- angle to the result.
	local snapAngle = math.pi / 2;
	vehicleRad = (math.floor((vehicleRad - bunkerRad) / snapAngle + 0.5) * snapAngle) + bunkerRad;
	dx, dz = MathUtil.getDirectionFromYRotation(vehicleRad);
	
	Debug.log("Snapped vehicleRad: " .. (vehicleRad) .. " dx: " .. dx .. " dz: " .. dz);
	
	-- Setup the DriveDirection and DriveTarget for the AIDriveStrategy.
    vehicleSpec.origDriveDirection = {dx, dz};
	vehicleSpec.currDriveDirection[1] = vehicleSpec.origDriveDirection[1];
	vehicleSpec.currDriveDirection[2] = vehicleSpec.origDriveDirection[2];
    local x,y,z = getWorldTranslation(vehicle:getAIDirectionNode())
    vehicleSpec.origDriveTarget = {x, z};
	vehicleSpec.currDriveTarget = vehicleSpec.origDriveTarget;
	
    local wx, wy, wz = worldDirectionToLocal(vehicleSpec.bunkerArea.startNode, dx, 0, dz);
    Debug.log("World direction: wx: " .. wx .. " wz: " .. wz);

    -- Set up the AINode that the worker uses as reference
    local AINode = clone(vehicleSpec.bunkerArea.startNode, true, false, false);
    vehicleSpec.bunkerArea.AINode = AINode;

    -- Rotate it so that its Z axis corresponds to vehicle Z axis
    setDirection(AINode, wx, wy, wz, 0, 1, 0);

    -- Move it, so that all coordinates for height/length is positive!
    x, _, z = 0,0,0;
    local hx, _, hz = localToLocal(vehicleSpec.bunkerArea.widthNode, AINode, 0, 0, 0);
    wx, _, wz = localToLocal(vehicleSpec.bunkerArea.heightNode, AINode, 0, 0, 0);

    if hx > 1 then
        -- This is our width, no need to translate x coordinate
        vehicleSpec.bunkerArea.width = hx;
    elseif hx < -1 then
        -- This is our width, but we need to move the AINode
        vehicleSpec.bunkerArea.width = math.abs(hx);
        x = hx;
    elseif wx > 1 then
        -- This is our width, no need to translate x coordinate
        vehicleSpec.bunkerArea.width = wx;
    elseif wx < -1 then
        -- This is our width, but we need to move the AINode
        vehicleSpec.bunkerArea.width = math.abs(wx);
        x = wx;
    end;

    if hz > 1 then
        -- This is our height, no need to translate z coordinate
        vehicleSpec.bunkerArea.height = hz;
    elseif hz < -1 then
        -- This is our width, but we need to move the AINode
        vehicleSpec.bunkerArea.height = math.abs(hz);
        z = hz;
    elseif wz > 1 then
        -- This is our width, no need to translate x coordinate
        vehicleSpec.bunkerArea.height = wz;
    elseif wz < -1 then
        -- This is our width, but we need to move the AINode
        vehicleSpec.bunkerArea.height = math.abs(wz);
        z = wz;
    end;

    setWorldTranslation(AINode, localToWorld(AINode, x, 0, z));

    Debug.log("BunkerArea Width: " .. vehicleSpec.bunkerArea.width .. " BunkerArea Height: " .. vehicleSpec.bunkerArea.height);
	
	-- Calculate vehicle dimensions
	CalculateVehicleBorders(vehicleSpec, vehicle);
		
	-- Reset any variables needed
    vehicleSpec.Direction = true;
	vehicleSpec.currSideShiftDirection = BunkerSiloWorker.SIDESHIFT_DIRECTION_LEFT;
	
	-- Record our starting pos so we can calculate minimum distance to travelled
	local tp = vehicleSpec.turnaroundpoint;
	tp.x, tp.y, tp.z = getWorldTranslation(vehicle.rootNode);
	
	-- Calculate our lookAheadDistance
	-- Use default of 3 if this is NOT an articulated machine
	vehicleSpec.lookAheadDistanceFront = 3;
	vehicleSpec.lookAheadDistanceRear = 3;
	
	-- Check if it is an articulated machine.
	if vehicle.spec_articulatedAxis ~= nil then
		if vehicle.spec_articulatedAxis.componentJoint ~= nil then
			-- This is an articulated machine, calculate the limits to make sure
			-- some of the articulated machines (like base-game loader) doesn't get
			-- crazy.
			
			local x, _, z = localToLocal(vehicle:getAISteeringNode(), vehicle.rootNode, 0, 0, 0);
			local dist = z; --MathUtil.vector2Length(x, z);

			Debug.log("dist: " .. dist);

			if dist + 3 < vehicleSpec.frontBorder then
				Debug.log("Compensating lookAheadFrontFront");
				vehicleSpec.lookAheadDistanceFront = vehicleSpec.frontBorder + math.abs(dist) + 1;
				Debug.log(vehicleSpec.lookAheadDistanceFront);
			else
				vehicleSpec.lookAheadDistanceFront = 3;
			end;

			if dist - 3 > vehicleSpec.rearBorder then	
				Debug.log("Compensating lookAheadRear");
				vehicleSpec.lookAheadDistanceRear = math.abs(vehicleSpec.rearBorder) + math.abs(dist) + 1;
				Debug.log(vehicleSpec.lookAheadDistanceRear);	
			else
				vehicleSpec.lookAheadDistanceRear = 3;
			end;
			
		end;
	end;

end

function CalculateVehicleBorders(vehicleSpec, vehicle)
	-- Set up the vehicle borders (with respect to the rootNode of the vehicle)
	-- TODO: Add check for attached implements also
	vehicleSpec.frontBorder =  vehicle.size.length/2 + vehicle.size.lengthOffset;
	vehicleSpec.rearBorder  = -vehicle.size.length/2 + vehicle.size.lengthOffset;
	vehicleSpec.leftBorder  =  vehicle.spec_aiDrivable.agentInfo.width/2;
	vehicleSpec.rightBorder = -vehicle.spec_aiDrivable.agentInfo.width/2;
	
	Debug.log("Calculated borders before wheels: " .. vehicleSpec.frontBorder .. " " .. vehicleSpec.rearBorder .. " " .. vehicleSpec.leftBorder .. " " .. vehicleSpec.rightBorder);
	
	-- Find out if our wheels are sticking out (add 0.1 as margin on each side)
	for _,wheel in pairs(vehicle.spec_wheels.wheels) do
		x1, y1, z1 = localToLocal(wheel.repr, vehicle.rootNode, -wheel.width/2, 0, 0);
		x2, y2, z2 = localToLocal(wheel.repr, vehicle.rootNode, wheel.width/2, 0, 0);
		
		-- Check if the left side of the wheel sticks out outside the tractor border.
		if math.sign(x1) > 0 and vehicleSpec.leftBorder < x1 + 0.1 then
			vehicleSpec.leftBorder = x1 + 0.1;
		elseif math.sign(x1) < 0 and vehicleSpec.rightBorder > x1 - 0.1 then
			vehicleSpec.rightBorder = x1 - 0.1;
		end
		
		-- Check if the right side of the wheel sticks out outside the tractor border.
		if math.sign(x2) > 0 and vehicleSpec.leftBorder < x2 + 0.1 then
			vehicleSpec.leftBorder = x2 + 0.1;
		elseif math.sign(x2) < 0 and vehicleSpec.rightBorder > x2 - 0.1 then
			vehicleSpec.rightBorder = x2 - 0.1;
		end
		
		-- Check for additional wheels connected to this "master" wheel (i.e. doubles)
		if wheel.additionalWheels ~= nil then
			for _,additionalWheel in pairs(wheel.additionalWheels) do
				local width = 0.5 * additionalWheel.width
				
				x1, y1, z1 = localToLocal(additionalWheel.wheelTire, vehicle.rootNode, -width, 0, 0);
				x2, y2, z2 = localToLocal(additionalWheel.wheelTire, vehicle.rootNode, width, 0, 0);
				
				if math.sign(x1) > 0 and vehicleSpec.leftBorder < x1 + 0.1 then
					vehicleSpec.leftBorder = x1 + 0.1;
				elseif math.sign(x1) < 0 and vehicleSpec.rightBorder > x1 - 0.1 then
					vehicleSpec.rightBorder = x1 - 0.1;
				end
				
				if math.sign(x2) > 0 and vehicleSpec.leftBorder < x2 + 0.1 then
					vehicleSpec.leftBorder = x2 + 0.1;
				elseif math.sign(x2) < 0 and vehicleSpec.rightBorder > x2 - 0.1 then
					vehicleSpec.rightBorder = x2 - 0.1;
				end
			end
		end
					
	end			
	
	Debug.log("Calculated borders after wheels: " .. vehicleSpec.frontBorder .. " " .. vehicleSpec.rearBorder .. " " .. vehicleSpec.leftBorder .. " " .. vehicleSpec.rightBorder);
end	

function AIDriveStrategyBunkerSilo:update(dt)
end

function AIDriveStrategyBunkerSilo:canDriveWithAttachedTool()
    -- Locate supported equipment
    if nil ~= self.vehicle.getAttachedImplements then
		local tools = self.vehicle:getAttachedImplements();
		if #tools > 0 then
			Debug.log("Unsupported tools attached!");
			return false;
		end
	end
	
	return true;
end

function AIDriveStrategyBunkerSilo:getDriveData(dt, vX, vY, vZ)

    local vehicle = self.vehicle
    local vehicleSpec = BunkerSiloWorker.getSpec(vehicle)

	-- First check that we still have a bunkersilo
	if vehicleSpec.bunkerSilo == nil or vehicleSpec.runState == BunkerSiloWorker.STATE_NONE then
		-- Stop the action
		return nil, nil, nil, 0, nil;
	end
	
	-- Depending on what state we are in...
	if vehicleSpec.runState == BunkerSiloWorker.STATE_WAITING then
		-- State is waiting
		vehicle:setBeaconLightsVisibility(true);

        local otherVehicleInBunker = false;

        for i=#vehicleSpec.bunkerSilo.vehicleList,1,-1 do
            if vehicleSpec.bunkerSilo.vehicleList[i] ~= vehicle then
                otherVehicleInBunker = true;
			end;
		end

		if not otherVehicleInBunker then
			Debug.log("Only me in bunker!");
			vehicleSpec.runState = BunkerSiloWorker.STATE_RUNNING;
		end

		return 0, 0, false, 0, 0;
		
	elseif vehicleSpec.runState == BunkerSiloWorker.STATE_RUNNING then

		vehicle:setBeaconLightsVisibility(false);		
		
        local otherVehicleInBunker = false;

        for i=#vehicleSpec.bunkerSilo.vehicleList,1,-1 do
            if vehicleSpec.bunkerSilo.vehicleList[i] ~= vehicle then
                otherVehicleInBunker = true;
			end;
		end

		-- Check if we got any other element in our list of
		--if #vehicleSpec.bunkerSilo.vehicleList > 1 then
        if otherVehicleInBunker then
			vehicleSpec.runState = BunkerSiloWorker.STATE_WAITING;

			if vehicle.isClient then
				if g_currentMission.player.farmId == vehicle.spec_aiJobVehicle.startedFarmId then
					local helperName = vehicle:getCurrentHelper().name
					local text = string.format(g_i18n:getText("BunkerSiloWorkerBlocked"), helperName)
					g_currentMission:addIngameNotification(FSBaseMission.INGAME_NOTIFICATION_CRITICAL, text)
				end
			end
								
			return 0, 0, false, 0, 0;
		end
		
		local tp = vehicleSpec.turnaroundpoint; -- The last turn-around position
		local cp = {};  -- Current position
		cp.x = 0;
		cp.y = 0;
		cp.z = 0;
		cp.x, cp.y, cp.z = getWorldTranslation(vehicle.rootNode);

		-- Calculate the distance we have travelled since last reversing the direction
		vehicleSpec.currTravelDistance = MathUtil.vector2Length(tp.x - cp.x, tp.z - cp.z);

		-- ### Find out if it is time to reverse direction ### --	
		local turnaround = false;

        -- Find out if we should use the fillLevel around us or not.
        local checkFillLevel = vehicleSpec.bunkerSilo.fillLevel > 10;

		--- Calculate some data for debug purposes
        if Debug.active then
            vehicleSpec.debugInfo.frontDistance = self:getFrontDistanceToBunkerEdge(vehicle, vehicleSpec);
            vehicleSpec.debugInfo.rearDistance = self:getRearDistanceToBunkerEdge(vehicle, vehicleSpec);
            vehicleSpec.debugInfo.leftDistance = self:getLeftDistanceToBunkerEdge(vehicle, vehicleSpec);
            vehicleSpec.debugInfo.rightDistance = self:getRightDistanceToBunkerEdge(vehicle, vehicleSpec);
        end

		local distanceToStop = 0;

		-- Check if we are moving forwards or backwards right now
		if vehicleSpec.Direction then
		
			-- Check if we still have chaff ahead
			vehicleSpec.fillLevel, vehicleSpec.areaChecked = self:getFillLevelOnFront(vehicle, vehicleSpec);
			
			-- Get distance to bunker edge
			vehicleSpec.currDistanceToBunkerEdge = self:getFrontDistanceToBunkerEdge(vehicle, vehicleSpec);
            distanceToStop = (vehicleSpec.currDistanceToBunkerEdge - vehicleSpec.safetyMarginFront);

			if (vehicleSpec.currDistanceToBunkerEdge < vehicleSpec.safetyMarginFront) or
			   ((vehicleSpec.fillLevel <= 0 and checkFillLevel) and vehicleSpec.currTravelDistance >= BunkerSiloWorker.MIN_TRAVEL_DISTANCE) then
				-- Turn direction
				turnaround = true;
			end
		else
			-- Check if we still have chaff behind us
			vehicleSpec.fillLevel, vehicleSpec.areaChecked = self:getFillLevelOnRear(vehicle, vehicleSpec);
			-- Get distance to bunker edge
			vehicleSpec.currDistanceToBunkerEdge = self:getRearDistanceToBunkerEdge(vehicle, vehicleSpec);
            distanceToStop = (vehicleSpec.currDistanceToBunkerEdge - vehicleSpec.safetyMarginRear);			

			if (vehicleSpec.currDistanceToBunkerEdge < vehicleSpec.safetyMarginRear) or
				((vehicleSpec.fillLevel <= 0 and checkFillLevel) and vehicleSpec.currTravelDistance >= BunkerSiloWorker.MIN_TRAVEL_DISTANCE) then
				-- Turn direction
				turnaround = true;
			end
		end

		-- Update necessary variables in case we should reverse direction.
		if turnaround == true then
			if vehicleSpec.Direction == true then
				vehicleSpec.Direction = false;
			else
				vehicleSpec.Direction = true;
				
                local sideShiftStep = BunkerSiloWorker.SIDESHIFT_STEP_SIZE;

				-- We now starts moving forwards, shift position to the side if possible
				if vehicleSpec.currSideShiftDirection == BunkerSiloWorker.SIDESHIFT_DIRECTION_LEFT then
					-- Check if we should continue this way
					local leftDistance = self:getLeftDistanceToBunkerEdge(vehicle, vehicleSpec);
					vehicleSpec.sideArea.fillLevel, vehicleSpec.sideArea.areaChecked = self:getFillLevelOnLeft(vehicle, vehicleSpec);
					
					if leftDistance - BunkerSiloWorker.SIDESHIFT_STEP_SIZE < vehicleSpec.safetyMarginSides then
                        sideShiftStep = leftDistance - vehicleSpec.safetyMarginSides;
                    end

                    if sideShiftStep < 0.2 or (vehicleSpec.sideArea.fillLevel <= 1 and checkFillLevel) then
						-- Time to switch direction
						vehicleSpec.currSideShiftDirection = BunkerSiloWorker.SIDESHIFT_DIRECTION_RIGHT;
                        sideShiftStep = BunkerSiloWorker.SIDESHIFT_STEP_SIZE;
					end
				else
					-- Check if we should continue this way
					local rightDistance = self:getRightDistanceToBunkerEdge(vehicle, vehicleSpec);
					vehicleSpec.sideArea.fillLevel, vehicleSpec.sideArea.areaChecked = self:getFillLevelOnRight(vehicle, vehicleSpec);
					
					if rightDistance - BunkerSiloWorker.SIDESHIFT_STEP_SIZE < vehicleSpec.safetyMarginSides then
                        sideShiftStep = rightDistance - vehicleSpec.safetyMarginSides;
                    end

					if sideShiftStep < 0.2 or (vehicleSpec.sideArea.fillLevel <= 1 and checkFillLevel) then
						-- Time to switch direction
						vehicleSpec.currSideShiftDirection = BunkerSiloWorker.SIDESHIFT_DIRECTION_LEFT;
                        sideShiftStep = BunkerSiloWorker.SIDESHIFT_STEP_SIZE;
					end
				end				

				local vRX, _,vRZ   = localDirectionToWorld(vehicle:getAISteeringNode(), 0,0,1);				
				vehicleSpec.currDriveTarget[1] = vehicleSpec.currDriveTarget[1] + vehicleSpec.currSideShiftDirection * sideShiftStep * vRZ;
				vehicleSpec.currDriveTarget[2] = vehicleSpec.currDriveTarget[2] + vehicleSpec.currSideShiftDirection * sideShiftStep * -vRX;
			end
			
			-- Record the position so we now how far we have travelled
			tp.x, tp.y, tp.z = getWorldTranslation(vehicle.rootNode);
		end

		-- Set currDriveDirection parameters depending on the desired travel direction
		if vehicleSpec.Direction == false then
			vehicleSpec.currDriveDirection[1] = vehicleSpec.origDriveDirection[1] * -1;
			vehicleSpec.currDriveDirection[2] = vehicleSpec.origDriveDirection[2] * -1;
		else
			vehicleSpec.currDriveDirection[1] = vehicleSpec.origDriveDirection[1];
			vehicleSpec.currDriveDirection[2] = vehicleSpec.origDriveDirection[2];	
		end

		-- ### Calculate new directions for the AI driver  ### --
		local pX, pZ = MathUtil.projectOnLine(vX, vZ, vehicleSpec.currDriveTarget[1], vehicleSpec.currDriveTarget[2], vehicleSpec.currDriveDirection[1], vehicleSpec.currDriveDirection[2])
		
		local tX, tZ;
		if vehicleSpec.Direction == false then
			tX = pX + vehicleSpec.currDriveDirection[1] * vehicleSpec.lookAheadDistanceRear; -- Put target 3 meters ahead (which seems suitable for most tractors)
			tZ = pZ + vehicleSpec.currDriveDirection[2] * vehicleSpec.lookAheadDistanceRear; -- Put target 3 meters ahead (which seems suitable for most tractors)
		else
			tX = pX + vehicleSpec.currDriveDirection[1] * vehicleSpec.lookAheadDistanceFront; -- Put target 3 meters ahead (which seems suitable for most tractors)
			tZ = pZ + vehicleSpec.currDriveDirection[2] * vehicleSpec.lookAheadDistanceFront; -- Put target 3 meters ahead (which seems suitable for most tractors)
		end;
		
        local maxSpeed = 15;

		vehicleSpec.debugInfo.pX = pX;
		vehicleSpec.debugInfo.pZ = pZ;
		vehicleSpec.debugInfo.tX = tX;
		vehicleSpec.debugInfo.tY = vY;
		vehicleSpec.debugInfo.tZ = tZ;
		vehicleSpec.debugInfo.vX = vX;
		vehicleSpec.debugInfo.vY = vY;
		vehicleSpec.debugInfo.vZ = vZ;

		return tX, tZ, vehicleSpec.Direction, maxSpeed, distanceToStop;
	end;

	-- Stop the action
	return nil, nil, nil, 0, nil;
end

function AIDriveStrategyBunkerSilo:getFrontDistanceToBunkerEdge(vehicle, spec)
	-- Transform vehicle position coordinates to bunker coordinates
	local vehicleX, _, vehicleZ = localToLocal(vehicle.rootNode, spec.bunkerArea.AINode, 0, 0, 0);
	spec.debugInfo.vehicleX = vehicleX;
	spec.debugInfo.vehicleZ = vehicleZ;

    return spec.bunkerArea.height - (vehicleZ + spec.frontBorder);
end

function AIDriveStrategyBunkerSilo:getRearDistanceToBunkerEdge(vehicle, spec)
	-- Transform vehicle position coordinates to bunker coordinates
	local vehicleX, _, vehicleZ = localToLocal(vehicle.rootNode, spec.bunkerArea.AINode, 0, 0, 0);
	spec.debugInfo.vehicleX = vehicleX;
	spec.debugInfo.vehicleZ = vehicleZ;	

    return vehicleZ + spec.rearBorder;
end

function AIDriveStrategyBunkerSilo:getLeftDistanceToBunkerEdge(vehicle, spec)
    -- Transform vehicle position coordinates to bunker coordinates
	local vehicleX, _, vehicleZ = localToLocal(vehicle.rootNode, spec.bunkerArea.AINode, 0, 0, 0);
	spec.debugInfo.vehicleX = vehicleX;
	spec.debugInfo.vehicleZ = vehicleZ;	
	
    return spec.bunkerArea.width - (vehicleX + spec.leftBorder);
end

function AIDriveStrategyBunkerSilo:getRightDistanceToBunkerEdge(vehicle, spec)
    -- Transform vehicle position coordinates to bunker coordinates
	local vehicleX, _, vehicleZ = localToLocal(vehicle.rootNode, spec.bunkerArea.AINode, 0, 0, 0);
	spec.debugInfo.vehicleX = vehicleX;
	spec.debugInfo.vehicleZ = vehicleZ;	

    return vehicleX + spec.rightBorder;
end

function AIDriveStrategyBunkerSilo:getFillLevelOnFront(vehicle, spec)
		
	--## Generate a square area, from the tractorPosition but in the direction of the silageBunker ##--
	
	-- Transform vehicle position coordinates to bunker coordinates
	local vehicleX, _, vehicleZ = localToLocal(vehicle.rootNode, spec.bunkerArea.AINode, 0, 0, 0);

    -- Create the square area
    spec.ftsX, _, spec.ftsZ = localToWorld(spec.bunkerArea.AINode, vehicleX + spec.rightBorder, 0, (vehicleZ + spec.frontBorder));
    spec.ftwX, _, spec.ftwZ = localToWorld(spec.bunkerArea.AINode, vehicleX + spec.leftBorder, 0, (vehicleZ + spec.frontBorder));
    spec.fthX, _, spec.fthZ = localToWorld(spec.bunkerArea.AINode, vehicleX + spec.rightBorder, 0, spec.bunkerArea.height);

    -- Get the fillLevel of the square area
	local _, level, totalArea = DensityMapHeightUtil.getFillLevelAtArea(spec.bunkerSilo.inputFillType,
																		spec.ftsX,
																		spec.ftsZ,
																		spec.ftwX,
																		spec.ftwZ,
																		spec.fthX,
																		spec.fthZ);
	return level, totalArea;
end

function AIDriveStrategyBunkerSilo:getFillLevelOnRear(vehicle, spec)

    --## Generate a square area, from the tractorPosition but in the direction of the silageBunker ##--
		
	-- Transform vehicle position coordinates to bunker coordinates
	local vehicleX, _, vehicleZ = localToLocal(vehicle.rootNode, spec.bunkerArea.AINode, 0, 0, 0);

    -- Create the square area
    spec.ftsX, _, spec.ftsZ = localToWorld(spec.bunkerArea.AINode, vehicleX + spec.rightBorder, 0, (vehicleZ + spec.rearBorder));
    spec.ftwX, _, spec.ftwZ = localToWorld(spec.bunkerArea.AINode, vehicleX + spec.leftBorder, 0, (vehicleZ + spec.rearBorder));
    spec.fthX, _, spec.fthZ = localToWorld(spec.bunkerArea.AINode, vehicleX + spec.rightBorder, 0, 0);

    -- Get the fillLevel of the square area
	local _, level, totalArea = DensityMapHeightUtil.getFillLevelAtArea(spec.bunkerSilo.inputFillType,
																		spec.ftsX,
																		spec.ftsZ,
																		spec.ftwX,
																		spec.ftsZ,
																		spec.ftsX,
																		spec.fthZ);

	return level, totalArea;
end

function AIDriveStrategyBunkerSilo:getFillLevelOnRight(vehicle, spec)
		
	--## Generate a square area on the right side of the tractor that fills the whole bunker area that is left ##--
	
	-- Transform vehicle position coordinates to bunker coordinates
	local vehicleX, _, _ = localToLocal(vehicle.rootNode, spec.bunkerArea.AINode, 0, 0, 0);

    -- Create the square area
    spec.sideArea.sX, _, spec.sideArea.sZ = localToWorld(spec.bunkerArea.AINode, 0, 0, 0);
    spec.sideArea.wX, _, spec.sideArea.wZ = localToWorld(spec.bunkerArea.AINode, vehicleX + spec.rightBorder, 0, 0);
    spec.sideArea.hX, _, spec.sideArea.hZ = localToWorld(spec.bunkerArea.AINode, 0, 0, spec.bunkerArea.height);
		
	-- Get the fillLevel within the area
	local _, level, totalArea = DensityMapHeightUtil.getFillLevelAtArea(spec.bunkerSilo.inputFillType,
																		spec.sideArea.sX,
																		spec.sideArea.sZ,
																		spec.sideArea.wX,
																		spec.sideArea.wZ,
																		spec.sideArea.hX,
																		spec.sideArea.hZ);

	return level, totalArea;
end

function AIDriveStrategyBunkerSilo:getFillLevelOnLeft(vehicle, spec)
		
	--## Generate a square area on the left side of the tractor that fills the whole bunker area that is left ##--
	local vehicleX, _, _ = localToLocal(vehicle.rootNode, spec.bunkerArea.AINode, 0, 0, 0);

    -- Create the square area
    spec.sideArea.sX, _, spec.sideArea.sZ = localToWorld(spec.bunkerArea.AINode, vehicleX + spec.leftBorder, 0, 0);
    spec.sideArea.wX, _, spec.sideArea.wZ = localToWorld(spec.bunkerArea.AINode, spec.bunkerArea.width, 0, 0);
    spec.sideArea.hX, _, spec.sideArea.hZ = localToWorld(spec.bunkerArea.AINode, vehicleX + spec.leftBorder, 0, spec.bunkerArea.height);
		
	-- Get the fillLevel within the area
	local _, level, totalArea = DensityMapHeightUtil.getFillLevelAtArea(spec.bunkerSilo.inputFillType,
																		spec.sideArea.sX,
																		spec.sideArea.sZ,
																		spec.sideArea.wX,
																		spec.sideArea.wZ,
																		spec.sideArea.hX,
																		spec.sideArea.hZ);														

	return level, totalArea;
end

function math.sign(x)
  return (x<0 and -1) or 1
end

Debug.log(("Script loaded: AIDriveStrategyBunkerSilo.lua - from %s (v%s)"):format(g_currentModName, g_modManager:getModByName(g_currentModName).version));
